#include "FGDependencyBetween.h"

#include "FGUtil.h"

#include "Constants.h"

#include <sstream>
#include <assert.h>

using namespace std;

namespace egstra {
  void fg_dependency_between::phi(const sentence* x, list<string> F[],
				  const int word_limit,
				  const bool use_pos) {
    if(!use_pos) { return; }
    const int N = x->size();
    ostringstream os;

    /* ctags[i] gives the coarse tag of the i'th token */
    string* ctags = new string[N];
    int* ictags = new int[N];
    /* vcnt[i] = # verbs from token 0 to token i */
    int* vcnt = new int[N];
    /* pcnt[i] = # punctuation tokens from token 0 to token i */
    int* pcnt = new int[N];
    /* ccnt[i] = # number of cc's from token 0 to token i */
    int* ccnt = new int[N];

    /* First traverse the sentence to precompute things */
    for(int i = 0 ; i < N; ++i) {
      const token& t = x->get_token(i);
      os.str("");
      os << t.coarse_tag();
      ctags[i] = os.str();
      ictags[i] = t.coarse_tag();
      if(i > 0) {
	vcnt[i] = vcnt[i - 1];
	pcnt[i] = pcnt[i - 1];
	ccnt[i] = ccnt[i - 1];
      } else {
	vcnt[i] = pcnt[i] = ccnt[i] = 0;
      }
      if(fg_util::is_verb(t))  { ++(vcnt[i]); }
      if(fg_util::is_punc(t))  { ++(pcnt[i]); }
      if(fg_util::is_coord(t)) { ++(ccnt[i]); }
    }

    /* we use these flags to ensure that we only generate each between
       feature once */
    bool* flags = new bool[fg_util::ntags()];
    for(int i = 0; i < fg_util::ntags(); ++i) { flags[i] = false; }

    /* Second, visit each dependency and generate its features */
    for(int s = 0; s < N; ++s) {
      /* generate features for all intervening tags */
      const string fl0 = "s+e+b:0+" + ctags[s] + "+";
      const string fl1 = "s+e+b:1+" + ctags[s] + "+";
      for(int e = s + 2; e < N; ++e) {
	const string fl0a = fl0 + ctags[e] + "+";
	const string fl1a = fl1 + ctags[e] + "+";
	/* create features, using the flags to check for uniqueness */
	list<string>& f_se = F[s + N*e];
	list<string>& f_es = F[e + N*s];
	for(int i = s + 1; i < e; ++i) {
	  const int ct = ictags[i];
	  if(flags[ct] == false) {
	    flags[ct] = true;
	    /* the "1" indicates that the modifier is before the head,
	       hence it belongs with (e,s) */
	    f_es.push_back(fl1a + ctags[i]);
	    f_se.push_back(fl0a + ctags[i]);
	  }
	}
	/* reset flags */
	for(int i = s + 1; i < e; ++i) { flags[ictags[i]] = false; }
      }

      /* generate features for counts of intervening classes */
      const string psp = "+" + ctags[s] + "+";
      for(int e = s + 1; e < N; ++e) {
	const string pspe = psp + ctags[e];
	list<string>& f_se = F[s + N*e];
	list<string>& f_es = F[e + N*s];

	/* number of verbs that appear between the two tokens
	   (i.e. from token s+1 to token e-1 both included) */
	int vcount = vcnt[e-1] - vcnt[s];
	if(vcount >= VERB_COUNT_THRESHOLD) vcount = VERB_COUNT_THRESHOLD;
	/* left-headed count */
	os.str("");
	os << "vcount:0+" << vcount;
	f_se.push_back(os.str());
	f_se.push_back(os.str() + pspe);
	/* right-headed count */
	os.str("");
	os << "vcount:1+" << vcount;
	f_es.push_back(os.str());
	f_es.push_back(os.str() + pspe);

	/* number of punctuation that appear between the two tokens
	   (i.e. from token s+1 to token e-1 both included) */
	int pcount = pcnt[e-1] - pcnt[s];
	if(pcount >= PUNC_COUNT_THRESHOLD) pcount = PUNC_COUNT_THRESHOLD;
	/* left-headed count */
	os.str("");
	os << "pcount:0+" << pcount;
	f_se.push_back(os.str());
	f_se.push_back(os.str() + pspe);
	/* right-headed count */
	os.str("");
	os << "pcount:1+" << pcount;
	f_es.push_back(os.str());
	f_es.push_back(os.str() + pspe);

	/* number of coordinations that appear between the two tokens
	   (i.e. from token s+1 to token e-1 both included) */
	int ccount = ccnt[e-1] - ccnt[s];
	if(ccount >= CC_COUNT_THRESHOLD) ccount = CC_COUNT_THRESHOLD;
	/* left-headed count */
	os.str("");
	os << "ccount:0+" << ccount;
	f_se.push_back(os.str());
	f_se.push_back(os.str() + pspe);
	/* right-headed count */
	os.str("");
	os << "ccount:1+" << ccount;
	f_es.push_back(os.str());
	f_es.push_back(os.str() + pspe);
      }
    }

    delete [] ctags;
    delete [] ictags;
    delete [] vcnt;
    delete [] pcnt;
    delete [] ccnt;

    delete [] flags;
  }
}
