/**
 * Implements the feature dictionary
 * @file FeatureDictionary.cc
 * @author Mihai Surdeanu
 */

#include <iostream>
#include <fstream>

#include "FeatureDictionary.h"
#include "CharUtils.h"
#include "Constants.h"
#include "CppAssert.h"

using namespace std;

namespace egstra {

  int FeatureDictionary::getFeature(const std::string & name, 
				    bool create)
  {
    StringMapEntry<IndexAndFrequency> * entry;

    // this feature already exists
    if((entry = mMap.get(name.c_str())) != NULL){
      // increment the feature frequency
      if(create) entry->value().mFreq ++;
      return entry->getValue().mIndex;
    }

    // the feature does not exist
    if(create == false) return -1;

    // must create a new feature
    // int index = mMap.size();
    ++mMaxDim;
    int index = mMaxDim;
    mMap.set(name.c_str(), IndexAndFrequency(index));
    //cerr << "Created feature \"" + name + "\" with index: " << index << "\n";
    return index;
  }

  void FeatureDictionary::load(const std::string & fileName)
  {
    // reset the previous content
    mMap.clear();
    mMaxDim = -1;
    int count = 0;  

    ifstream is(fileName.c_str());
    RVASSERT(is, "Can not open file: " << fileName);
    char line[MAX_FEATDICT_LINE];
  
    while(is.getline(line, MAX_FEATDICT_LINE)){
      vector<string> tokens;
      simpleTokenize(line, tokens, " \t");
      RVASSERT(tokens.size() == 3, "Invalid dictionary entry: " << line);
      int freq = toInteger(tokens[2]);
      int index = toInteger(tokens[1]);
    
      if(freq >= UNKNOWN_FEATURE_THRESHOLD){
	mMap.set(tokens[0].c_str(), IndexAndFrequency(index, freq));
	if (index>mMaxDim) {
	  mMaxDim = index;
	}
	count ++;
      }
    }

    is.close();
    cerr << "FeatureDictionary :: loaded " << count << " features from " << fileName << ", max dimension is " << mMaxDim << "\n";
  }

  void FeatureDictionary::save(const std::string & fileName)
  {
    ofstream os(fileName.c_str());
    RVASSERT(os, "Can not create file: " << fileName);
    int count = 0;

    vector<string> keys;
    mMap.getKeys(keys);

    for(size_t i = 0; i < keys.size(); i ++){
      IndexAndFrequency iaf;
      mMap.get(keys[i].c_str(), iaf);
      os << keys[i] << " " << iaf.mIndex << " " << iaf.mFreq << "\n";
      count ++;
    }

    os.close();
    cerr << "Saved " << count << " features to " << fileName << "\n";
  }

  void FeatureDictionary::map_all(vector<int>& fidx,
				  const list<string>& fstr,
				  const bool create) {
    assert(fidx.size() == 0);
    fidx.reserve(fstr.size());
    list<string>::const_iterator it = fstr.begin();
    const list<string>::const_iterator it_end = fstr.end();
    for(; it != it_end; ++it) {
      const int idx = getFeature(*it, create);
      assert(idx >= -1);
      if(idx >= 0) { fidx.push_back(idx); }
    }
  }

  int FeatureDictionary::map_all(int* const fidx,
				 const list<string>& fstr,
				 const bool create) {
    list<string>::const_iterator it = fstr.begin();
    const list<string>::const_iterator it_end = fstr.end();
    int n = 0;
    for(; it != it_end; ++it) {
      const int idx = getFeature(*it, create);
      assert(idx >= -1);
      if(idx >= 0) { fidx[n] = idx; ++n; }
    }
    return n;
  }
}
