#include "ObjectiveFunction.h"

#include "OptUtils.h"

#include <stdlib.h>

#include <math.h>
#include <iomanip>
#include <iostream>

using namespace std;

namespace egstra {

  /***** Log-Linear objective *****/

  /* brief identifier */
  const char* obj_loglinear::name = "LL";

  /* compute the log-likelihood using the partition function.  assert
     that the log-likelihood is within a tolerance of zero. */
  double obj_loglinear::exprimal(const example* const ex,
				 double* const S,
				 structured_model& M) {
    const pattern* const x = ex->x();
    const label& y = ex->y();

    /* compute the true label's score */
    double gold_score = 0;
    label::const_iterator it = y.begin();
    const label::const_iterator it_end = y.end();
    for(; it != it_end; ++it) { gold_score += S[*it]; }

    /* log P(y|x;S) = S(gold tree) - log Z(S) */
    const double logZ = M.partition(x, S);
    const double gold_logp = gold_score - logZ;

    /* check the consistency of the log-probability */
    assert(isfinite(gold_logp));
    if(gold_logp > 0) {
      if(gold_logp > EGSTRA_OBJECTIVEFUNCTION_PRIMAL_TOL/2) {
	cerr << "*** warning: gold_logp = " << scientific << setprecision(15)
	     << gold_logp << " at pattern " << x->id()
	     << " (N=" << x->size() << " R=" << x->nparts() << ")"
	     << endl;
      }
      assert(gold_logp <= EGSTRA_OBJECTIVEFUNCTION_PRIMAL_TOL);
      /* truncate the log-probability to zero */
      return 0;
    }
    return gold_logp;
  }

  double obj_loglinear::exgrad(parameters& g,
			       const double scale,
			       const example* const ex,
			       double* const S,
			       const struct fvec* const F,
			       structured_model& M) {
    const pattern* const x = ex->x();
    const int R = x->nparts();

    /* compute the marginal probabilities */
    double* const P = new double[R];
    const double log_Z = M.marginals(x, S, P);

    /* create an array indicating which parts are "correct" */
    char* const correct = optutils::get_correct(ex);

    /* compute (empirical - expected) and gold-logp */
    double gold_score = 0;
    for(int r = 0; r < R; ++r) {
      double empirical = 0.0;
      if(correct[r]) {
	gold_score += S[r];
	empirical = 1.0;
      }
      const double expected = P[r];
      const double mul = scale*(empirical - expected);
      g.add(F + r, mul);
    }

    free(correct);
    delete [] P;

    /* check the consistency of the log-probability */
    const double gold_logp = gold_score - log_Z;
    assert(isfinite(gold_logp));
    if(gold_logp > 0) {
      if(gold_logp > EGSTRA_OBJECTIVEFUNCTION_PRIMAL_TOL/2) {
	cerr << "*** warning: gold_logp = " << scientific << setprecision(15)
	     << gold_logp << " at pattern " << x->id()
	     << " (N=" << x->size() << " R=" << x->nparts() << ")"
	     << endl;
      }
      assert(gold_logp <= EGSTRA_OBJECTIVEFUNCTION_PRIMAL_TOL);
      /* truncate the log-probability to zero */
      return 0;
    }

    /* the objective is scaled */
    return scale*gold_logp;
  }


  /***** Max-Margin objective *****/

  /* brief identifier */
  const char* obj_maxmargin::name = "MM";
  /* scaling for the Hamming loss within the max-margin objective */
  double obj_maxmargin::loss_scale = 1.0;

  /* explicitly compute the hinge-loss by performing a viterbi search
     on adjusted scores (note that this alters S).  note that the
     hinge loss is

     max_y {Loss(y) - Margin(y)}
       = max_y {Loss(y) + W . f(x_i,y)} - W . f(x_i,y_i)

     this corresponds to defining a part score vector S as follows

     S(part) = Loss(part) + W . f(x_i,part)

     and computing the score of the viterbi tree arising from S, and
     finally subtracting the gold tree's score. */
  double obj_maxmargin::exprimal(const example* const ex,
				 double* const S,
				 structured_model& M) {
    const pattern* const x = ex->x();
    const int R = x->nparts();

    /* adjust the part scores to compute the hinge loss. */
    char* const correct = optutils::get_correct(ex);
    double gold_score = 0;
    for(int r = 0; r < R; ++r) {
      if(correct[r]) { gold_score += S[r]; }
      else { S[r] += loss_scale; }
    }
    free(correct);

    /* get the worst margin violator and compute its score */
    label yhat;
    const double hat_score = M.argmax(x, S, yhat);

    const double hinge_loss = hat_score - gold_score;
    assert(isfinite(hinge_loss));
    if(hinge_loss < 0) {
      if(hinge_loss < -EGSTRA_OBJECTIVEFUNCTION_PRIMAL_TOL/2) {
	cerr << "*** warning: hinge_loss = " << scientific << setprecision(15)
	     << hinge_loss << " at pattern " << x->id()
	     << " (N=" << x->size() << " R=" << x->nparts() << ")"
	     << endl;
      }
      assert(hinge_loss >= -EGSTRA_OBJECTIVEFUNCTION_PRIMAL_TOL);
      /* truncate the hinge-loss to zero */
      return 0;
    }

    /* return the *negative* hinge loss, because we frame the primal
       as a maximization problem */
    return -hinge_loss;
  }

  double obj_maxmargin::exgrad(parameters& g,
			       const double scale,
			       const example* const ex,
			       double* const S,
			       const struct fvec* const F,
			       structured_model& M) {
    const pattern* const x = ex->x();
    const label& y = ex->y();

    const int R = x->nparts();

    /* adjust the part scores to compute the hinge loss. */
    char* const correct = optutils::get_correct(ex);
    double gold_score = 0;
    for(int r = 0; r < R; ++r) {
      if(correct[r]) { gold_score += S[r]; }
      else { S[r] += loss_scale; }
    }
    free(correct);

    /* get the worst margin violator and compute its score */
    label yhat;
    const double hat_score = M.argmax(x, S, yhat);

    const double hinge_loss = hat_score - gold_score;
    assert(isfinite(hinge_loss));
    if(hinge_loss < 0) {
      if(hinge_loss < -EGSTRA_OBJECTIVEFUNCTION_PRIMAL_TOL/2) {
	cerr << "*** warning: hinge_loss = " << scientific << setprecision(15)
	     << hinge_loss << " at pattern " << x->id()
	     << " (N=" << x->size() << " R=" << x->nparts() << ")"
	     << endl;
      }
      assert(hinge_loss >= -EGSTRA_OBJECTIVEFUNCTION_PRIMAL_TOL);
      /* truncate the hinge-loss to zero.  note that returning here is
	 fine, since when the hinge loss is zero, the gradient is
	 zero. */
      return 0;
    }

    /* find each label's contribution to the gradient:
       f(xi,yi) - f(xi,y) */
    double* const gcontrib = (double*)calloc(R, sizeof(double));
    label::const_iterator it = y.begin();
    label::const_iterator it_end = y.end();
    for(; it != it_end; ++it) { gcontrib[*it] += scale; }
    it = yhat.begin();
    it_end = yhat.end();
    for(; it != it_end; ++it) { gcontrib[*it] -= scale; }

    /* now perform the updates.  we skip updates that are zero or
       close enough to zero (note that the contributions will always
       be in units of "scale"). */
    const double thresh = (scale < 0 ? -scale/2 : scale/2);
    for(int r = 0; r < R; ++r) {
      const double gc = gcontrib[r];
      const double gcabs = (gc < 0 ? -gc : gc);
      if(gcabs > thresh) { g.add(F + r, gc); }
    }

    free(gcontrib);

    /* the negative hinge loss is scaled */
    return -scale*hinge_loss;
  }
}
