#include "OptUtils.h"

#include "Timer.h"
#include "Constants.h"

#include <stdio.h>
#include <stdlib.h>

#include <iostream>

/* detailed name of the trainer */
#define myname								\
  "OptUtils[" << M.name() << "]"

using namespace std;

namespace egstra {

  char* optutils::get_correct(const example* const ex) {
    const pattern* const x = ex->x();
    const int R = x->nparts();
    char* correct = (char*)calloc(R, sizeof(char));
    const label& y = ex->y();
    label::const_iterator it = y.begin();
    const label::const_iterator it_end = y.end();
    for(; it != it_end; ++it) {
      const int part = *it;
      assert(part >= 0);
      assert(part < R);
      correct[part] = 1;
    }
    return correct;
  }

  int optutils::count_matches(const example* const ex,
			      const label& yhat) {
    const int R = ex->x()->nparts();
    char* const correct = get_correct(ex);
    /* now count the number of matches */
    int match = 0;
    label::const_iterator yi = yhat.begin();
    const label::const_iterator yend = yhat.end();
    for(; yi != yend; ++yi) {
      const int part = *yi;
      assert(part >= 0);
      assert(part < R);
      match += correct[part];
    }
    free(correct);
    return match;
  }

  void optutils::classify_val_set(const parameters& V,
				  const dataset& data,
				  fgen& fg,
				  structured_model& M,
				  const string& dir,
				  const int t,
				  const bool minrisk) {
    cerr << myname << " : decoding validation set : " << flush;
    timer_start();

    /* open two output files */
    char buf[64];
    sprintf(buf, "max-labels.%03d", t);
    string maxfilename = dir + "/" + buf;
    sprintf(buf, "mbr-labels.%03d", t);
    string mbrfilename = dir + "/" + buf;
    FILE* maxout = fopen(maxfilename.c_str(), "w");
    assert(maxout != NULL);
    FILE* mbrout = NULL;
    if(minrisk) {
      mbrout = fopen(mbrfilename.c_str(), "w");
      assert(mbrout != NULL);
    }

    /* use these to compute the part-wise precision and recall */
    int correct_parts = 0;
    int max_matching_parts = 0;
    int max_predicted_parts = 0;
    int mbr_matching_parts = 0;
    int mbr_predicted_parts = 0;

    const int N = data.size();
    int cnt = 0;
    dataset::const_iterator it = data.begin();
    const dataset::const_iterator it_end = data.end();
    for(; it != it_end; ++it) {
      const example* ex = *it;
      const pattern* const x = ex->x();

      /* compute inner products */
      const int R = x->nparts();
      double* const S = (double*)malloc(R*sizeof(double));
      const struct fvec* F = fg.phi(x);
      V.dot(S, F, R);
      delete [] F;

      const label& y = ex->y();
      correct_parts += y.size();

      /* find the argmax (viterbi) label */
      label ymax;
      M.argmax(x, S, ymax);
      M.fwrite(maxout, x, ymax);
      max_predicted_parts += ymax.size();
      max_matching_parts += count_matches(ex, ymax);

      /* find the min-risk label */
      if(minrisk) {
	double* const P = (double*)malloc(R*sizeof(double));
	M.marginals(x, S, P);
	label ymbr;
	M.argmax(x, P, ymbr);
	free(P);
	M.fwrite(mbrout, x, ymbr);
	mbr_predicted_parts += ymbr.size();
	mbr_matching_parts += count_matches(ex, ymbr);
      }

      free(S);

      ++cnt;
      ticker(cnt, N);
    }
    ticker(N, N);
    fclose(maxout);
    if(minrisk) { fclose(mbrout); }
    timer_stop();
    cerr << " done [" << timestr(timediff) << "]" << endl;
    assert(correct_parts > 0);
    assert(max_predicted_parts > 0);
    const double maxP =
      100.0*(double)max_matching_parts/(double)max_predicted_parts;
    const double maxR =
      100.0*(double)max_matching_parts/(double)correct_parts;
    cerr << myname << " : argmax prec/rec = "
	 << fixed << setprecision(2)
	 << maxP << " " << maxR << endl;
    if(minrisk) {
      assert(mbr_predicted_parts > 0);
      const double mbrP =
	100.0*(double)mbr_matching_parts/(double)mbr_predicted_parts;
      const double mbrR =
	100.0*(double)mbr_matching_parts/(double)correct_parts;
      cerr << myname << " : min-risk prec/rec = "
	   << fixed << setprecision(2)
	   << mbrP << " " << mbrR << endl;
    }
  }


  void optutils::load_datasets(structured_model& M,
			       const char* const trfile,
			       dataset& trdata,
			       const char* const valfile,
			       dataset& valdata) {
    cerr << myname << " : loading datasets... " << flush;
    int id = 0;
    pattern* x = NULL;
    label* y = NULL;

    assert(trdata.size() == 0);
    assert(valdata.size() == 0);

    /* load the training set */
    FILE* in = fopen(trfile, "r");
    assert(in != NULL);
    while(M.fread(in, x, y)) {
      x->set_id(id);
      ++id;
      trdata.push_back(new example(x, y));
      x = NULL;
      y = NULL;
    }
    fclose(in);
    cerr << trdata.size() << " tr " << flush;

    /* load validation set if provided */
    if(valfile != NULL) {
      /* NB: we do not reset the id counter; thus all patterns have
	 unique ids. */
      in = fopen(valfile, "r");
      assert(in != NULL);
      while(M.fread(in, x, y)) {
	x->set_id(id);
	++id;
	valdata.push_back(new example(x, y));
	x = NULL;
	y = NULL;
      }
      fclose(in);
      cerr << valdata.size() << " val" << flush;
    }
    cerr << endl;
  }

  void optutils::clear_dataset(dataset& data) {
    const int N = data.size();
    for(int i = 0; i < N; ++i) { delete data[i]; }
    data.clear();
  }


  const example* optutils::randselect(const int rand_type,
				      const int i,
				      dataset& data) {
    const int N = data.size();
    /* produce a random value in [0, 1) */
    const double rand_01 = ((double)random())/(1.0 + (double)RAND_MAX);
    if(rand_type == EGSTRA_RAND_UNIFORM) {
      /* in this case, just pick an example at random */
      const double rand_0n = N*rand_01;
      const int rand_idx = (int)rand_0n;
      return data[rand_idx];
    } else if(rand_type == EGSTRA_RAND_PERMUTATION) {
      /* in this case, compute a random permutation on-the-fly by
	 picking a random example out of [i, n) and swapping it
	 with the current example (i.e., the Knuth shuffle). */
      const double rand_in = i + (N - i)*rand_01;
      const int rand_idx = (int)rand_in;
      /* swap and store the random selection in x */
      const example* ex = data[rand_idx];
      data[rand_idx] = data[i];
      data[i] = ex;
      return ex;
    } else if(rand_type == EGSTRA_DETERMINISTIC) {
      /* deterministic ascending order */
      return data[i];
    }

    cerr << "unrecognized randomization type "
	 << rand_type << endl;
    assert(0);

    return NULL;
  }
}

#undef myname
