/************************************************************
 *
 *   Author : Terry Koo
 *            maestro@csail.mit.edu
 *
 *   Class containing routines for computing a regularized
 *   primal objective/gradient.  The objective functions are
 *   of the form
 *
 *   f(w; S) = (C/2) ||w||^2 + (1/n) \sum_i loss(x_i, y_i; w)
 *
 ************************************************************/
#ifndef EGSTRA_REG_PRIMAL_H
#define EGSTRA_REG_PRIMAL_H

#include "FGen.h"
#include "StructuredModel.h"
#include "Parameters.h"
#include "Dataset.h"

#include <vector>

namespace egstra {

  /**
   * RegPrimal : a templatized class dependent on an objective
   * function Obj (cf. ObjectiveFunction.{cc,h}, REGTrainer.{cc,h})
   *
   * Obj is expected to provide the following meathods
   *
   *   Obj::exprimal(example* ex, double* S, structured_model M)
   *   Obj::exgrad(parameters g, double scale, example* ex,
   *               double* S, fvec* F, structured_model M)
   *     These return the example-wise contribution to the primal
   *     objective and the primal gradient
   **/
  template<class Obj>
  class reg_primal {
  public:

    /* given the training set and primals, compute the regularized
       primal objective: (C/2)||V||^2 + \sum_i Loss(x_i, y_i; V) */
    static double objective(parameters& V,
			    const double C,
			    const dataset& data,
			    fgen& fg,
			    structured_model& M,
			    const double scale = 1.0);

    /* given the training set and primals, compute the gradient of the
       regularized primal objective: g = C*V + \sum_i dLoss(x_i,y_i)/dw.
       in addition, return the primal objective. */
    static double gradient(parameters& g,
			   parameters& V,
			   const double C,
			   const dataset& data,
			   fgen& fg,
			   structured_model& M,
			   const double scale = 1.0);
  };

}

#include "RegPrimal.tcc"


#endif /* EGSTRA_REG_PRIMAL_H */
