#include "RegPrimal.h"

#include "Timer.h"

#include <iostream>

/* detailed name of the trainer */
#define myname								\
  "RegPrimal[" << M.name() << "," << Obj::name << "]"

using namespace std;

namespace egstra {

  template<class Obj>
  double reg_primal<Obj>::objective(parameters& V,
				    const double C,
				    const dataset& data,
				    fgen& fg,
				    structured_model& M,
				    const double scale) {
    cerr << myname << " : primal objective : " << flush;
    timer_start();

    dataset::const_iterator it = data.begin();
    const dataset::const_iterator it_end = data.end();
    const int N = data.size();
    int cnt = 0;
    double primal = 0;
    for(; it != it_end; ++it) {
      const example* ex = *it;
      const pattern* const x = ex->x();
      /* compute inner products */
      const int R = x->nparts();
      double* const S = new double[R];
      const struct fvec* F = fg.phi(x);
      V.dot(S, F, R);
      /* accumulate objective */
      primal += Obj::exprimal(ex, S, M);
      delete [] S;
      delete [] F;
      ++cnt;
      ticker(cnt, N);
    }
    ticker(N, N);
    timer_stop();
    cerr << " done [" << timestr(timediff) << "]" << endl;

    /* force explicit recomputation of V^2*/
    primal = scale*(primal/N - 0.5*C*V.sqL2(true));

    return primal;
  }


  template<class Obj>
  double reg_primal<Obj>::gradient(parameters& g,
				   parameters& V,
				   const double C,
				   const dataset& data,
				   fgen& fg,
				   structured_model& M,
				   const double scale) {
    cerr << myname << " : primal gradient : " << flush;
    timer_start();

    dataset::const_iterator it = data.begin();
    const dataset::const_iterator it_end = data.end();
    const int N = data.size();

    /* initialize the gradient with -C*V */
    g.set_values(V, -scale*C);
    const double exscale = scale/N;
    int cnt = 0;
    double primal = 0;
    for(; it != it_end; ++it) {
      const example* ex = *it;
      const pattern* const x = ex->x();
      /* compute inner products */
      const int R = x->nparts();
      double* const S = new double[R];
      const struct fvec* F = fg.phi(x);
      V.dot(S, F, R);
      /* accumulate gradient */
      primal += Obj::exgrad(g, exscale, ex, S, F, M);
      delete [] S;
      delete [] F;
      ++cnt;
      ticker(cnt, N);
    }
    ticker(N, N);
    timer_stop();
    cerr << " done [" << timestr(timediff) << "]" << endl;

    /* force explicit recomputation of V^2 */
    primal = primal - 0.5*scale*C*V.sqL2(true);

    return primal;
  }

}

#undef myname
