/**
 * Implements a generic hash map with string keys
 * @file StringMap.h
 * @author Mihai Surdeanu
 */

#include <list>

#include "HashMap.h"
#include "CharArrayEqualFunc.h"
#include "CharArrayHashFunc.h"

#ifndef EGSTRA_STRING_MAP_H
#define EGSTRA_STRING_MAP_H

namespace egstra {

  template <class T>
  class StringMapEntry {

  public:
    StringMapEntry(char * key,
		   const T & value) {
      _key = key;
      _value = value;
    };

    char * getKey() const { return _key; };

    const T & getValue() const { return _value; };
    T & value() { return _value; }

    void setValue(const T & v) { _value = v; };

    ~StringMapEntry() {
      delete [] _key;
    };

  private:
    char * _key;
    T _value;
  };

  template <class T>
  class StringMap {

  public:

    typedef typename egstra::HashMap<const char *, StringMapEntry<T> *,
      egstra::CharArrayHashFunc, egstra::CharArrayEqualFunc>::iterator 
      iterator;

    typedef typename egstra::HashMap<const char *, StringMapEntry<T> *,
      egstra::CharArrayHashFunc, egstra::CharArrayEqualFunc>::const_iterator 
      const_iterator;

    typedef typename std::list<StringMapEntry<T> *>::iterator
      entry_list_iterator;

    StringMap() {
    }

    ~StringMap() {
      clear();
    }

    void clear() {
      //std::cerr << "StringMap::clear() begin\n";

      // fetch all entry pointers
      std::list<StringMapEntry<T> *> entries;
      for(iterator it = _map.begin(); it != _map.end(); it ++){
	entries.push_back((* it).second);
      }

      // clear the map
      _map.clear();

      // delete all entries
      for(entry_list_iterator it = entries.begin(); 
	  it != entries.end(); it ++){
	delete (* it);
      }

      //std::cerr << "StringMap::clear() end\n";
    }

    bool set(const char * key, const T & value) {
      if(contains(key)){
	return false;
      }
      
      int keyLen = 0;
      for(; key[keyLen] != 0; keyLen ++);
      
      char * newKey = new char[keyLen + 1];
      for(int i = 0; i < keyLen; i ++) newKey[i] = key[i];
      newKey[keyLen] = '\0';
      
      StringMapEntry<T> * p = new StringMapEntry<T>(newKey, value);
      _map[p->getKey()] = p;
      
      return true;
    }

    bool overwrite(const char * key, const T & value) {
      if(contains(key)){
	iterator it = _map.find(key);
	(* it).second->setValue(value);
	return true;
      } else {
	return set(key, value);
      }
    }

    bool get(const char * key, T & value) const {
      const_iterator it;
      
      if((it = _map.find(key)) != _map.end()){
	value = (* it).second->getValue();
	return true;
      }
      
      return false;
    }

    StringMapEntry<T> * get(const char * key) const {
      const_iterator it;
      
      if((it = _map.find(key)) != _map.end()){
	return (* it).second;
      }
      
      return NULL;
    }

    bool contains(const char * key) const {
      if(_map.find(key) != _map.end()){
	return true;
      }
      
      return false;
    }

    void getKeys(std::vector<std::string> & keys) const {
      for(const_iterator it = begin(); it != end(); it ++){
	keys.push_back((* it).second->getKey());
      }
    }

    size_t size() const { return _map.size(); };

    bool empty() const { return _map.empty(); };

    const_iterator begin() const { return _map.begin(); };

    const_iterator end() const { return _map.end(); };

    iterator mbegin() { return _map.begin(); };

    iterator mend() { return _map.end(); };

  protected:

    egstra::HashMap<const char *, StringMapEntry<T> *,
      egstra::CharArrayHashFunc,
      egstra::CharArrayEqualFunc> _map;

  };

  void loadStringIntegerMap(StringMap<int> & map,
			    const std::string & fileName);
}

#endif /* EGSTRA_STRING_MAP_H */
