#include "DepFeats1.h"


/* The part-numbering for labeled dependencies is matched defined as
   follows:

   part-index(h,m,l) = h + N*m + N*N*l

   where h is the head word index, m is the modifier word index, l is
   the label index, and N is the length of the sentence (excluding
   root token).  Root dependencies are numbered with h == m.  Note
   that this matches the part-numbering of the first-order dependency
   parsers. */


#include "CharUtils.h"
#include "Options.h"

/* subordinate feature generation functions */
#include "FGUtil.h"
#include "FGToken.h"
#include "FGRootToken.h"
#include "FGTokenContext.h"
#include "FGDependency.h"
#include "FGDependencyContext.h"
#include "FGDependencyBetween.h"
#include "FGDependencyDistance.h"

#include <iostream>
#include <assert.h>
#include <stdlib.h>

using namespace std;

#define CONF_DEFAULT "t|tc|d|dc|b|dist|+w|+p"

namespace egstra {
  void depfeats1::usage(const char* const mesg) {
    cerr << _name << " options:" << endl;
    cerr << " --generate    : feature-generation mode" << endl;
    cerr << " --L=<int>     : number of dependency labels" << endl;
    cerr << " --fdict=<str> : pathname to feature-dictionary directory" << endl;
    cerr << " --fconf=\"<flag>|...\" : feature-configuration flags" << endl;
    cerr << "     t    : token features" << endl;
    cerr << "     tc   : token context features" << endl;
    cerr << "     d    : dependency features" << endl;
    cerr << "     dc   : dependency context features" << endl;
    cerr << "     b    : depenedency between features" << endl;
    cerr << "     dist : distance features" << endl;
    cerr << "     +w   : use all word information" << endl;
    cerr << "     w<N> : use only the top-N most frequent words" << endl;
    cerr << "     +p   : use POS information" << endl;
    cerr << "     ~p   : don't use POS information" << endl;
    cerr << "   (default \"" << CONF_DEFAULT << "\")" << endl;
    cerr << endl;
    cerr << mesg << endl;
  }

  void depfeats1::process_options() {
    int tmp;
    _K = -1;
    if(!options::get("L", _K)) {
      usage("please provide the number of dependency labels"); exit(1);
    }

    _generation_mode = false;
    if(options::get("generate", tmp)) {
      _generation_mode = true;
    }

    string dict;
    if(!options::get("fdict", dict)) {
      usage("please provide a feature dictionary"); exit(1);
    }
    _dictdir = dict;

    string conf;
    if(!options::get("fconf", conf)) {
      cerr << _name << " : using default feature configuration \""
	   << CONF_DEFAULT << "\"" << endl;
      conf = CONF_DEFAULT;
    }

    _use_token = false;
    _use_token_context = false;
    _use_dependency = false;
    _use_dependency_context = false;
    _use_dependency_between = false;
    _use_dependency_distance = false;

    _word_limit = -1;
    _use_pos = false;

    cerr << _name << " : feature conf \"" << conf << "\"" << endl;

    vector<string> flags;
    simpleTokenize(conf, flags, "|");

    for (vector<string>::const_iterator i = flags.begin();
	 i != flags.end();
	 ++i) {
      if (*i == "t") {
	cerr << _name << " : activating token features\n";
	_use_token = true;
      } else if (*i == "tc") {
	cerr << _name << " : activating token context features\n";
	_use_token_context = true;
      } else if (*i == "d") {
	cerr << _name << " : activating dependency features\n";
	_use_dependency = true;
      } else if (*i == "dc") {
	cerr << _name << " : activating dependency context features\n";
	_use_dependency_context = true;
      } else if (*i == "dist") {
	cerr << _name << " : activating distance features\n";
	_use_dependency_distance = true;
      } else if (*i == "b") {
	cerr << _name << " : activating between features\n";
	_use_dependency_between = true;
      } else if (*i == "+w") {
	cerr << _name << " : using word information\n";
	_word_limit = -1;
      } else if (i->c_str()[0] == 'w') {
	_word_limit = atoi(i->c_str() + 1);
	cerr << _name << " : using only top-" << _word_limit << " words\n";
      } else if (*i == "+p") {
	cerr << _name << " : using POS information\n";
	_use_pos = true;
      } else if (*i == "~p") {
	cerr << _name << " : ignoring POS information\n";
	_use_pos = false;
      } else {
	string str = "unknown feature flag \"" + (*i) + "\"";
	usage(str.c_str()); exit(1);
      }
    }

    load_dictionaries();
    compute_dim();
  }


  depfeats1::~depfeats1() {
    /* save the feature dictionaries (only if they were generated) */
    save_dictionaries();
  }

  void depfeats1::compute_dim() {
    _unlab_dim = 0;
    if(_use_token || _use_token_context) {
      /* dimensions for the head and for the modifier */
      _unlab_dim += 2*(_tok_dict.dimensionality());
    }
    if(_use_dependency ||
       _use_dependency_context ||
       _use_dependency_distance ||
       _use_dependency_between) {
      _unlab_dim += _dep_dict.dimensionality();
    }
    _dim = _unlab_dim*_K;
  }


  void depfeats1::load_dictionaries() {
    string path = _dictdir;
    if(!_generation_mode) {
      cerr << _name << " : loading feature dictionaries from \""
	   << _dictdir << "\"" << endl;
      _tok_dict.load(path + "/token.features");
      _dep_dict.load(path + "/dependency.features");
    }
    fg_util::load_tag_map(path + "/tags.map");
  }

  void depfeats1::save_dictionaries() {
    if(_generation_mode) {
      cerr << _name << " : saving feature dictionaries to \""
	   << _dictdir << "\"" << endl;
      string path = _dictdir;
      _tok_dict.save(path + "/token.features");
      _dep_dict.save(path + "/dependency.features");
    }
  }


  const struct fvec* depfeats1::phi(const pattern* const x) {
    const int N = x->size();
    const int NN = N*N;
    const int NNK = NN*_K;
    assert(NNK == x->nparts());
    const sentence* sent_x = (const sentence*)x;

    /* token features */
    vector<int>* tok_feats = new vector<int>[N];
    vector<int> root_feats;
    /* generate the features by first creating strings and then
       mapping them to integers */
    {
      list<string>* tok_feats_str = new list<string>[N];
      list<string> root_feats_str;
      if(_use_token) {
	fg_token::phi(sent_x, tok_feats_str,
		      _word_limit, _use_pos);
	fg_root_token::phi(root_feats_str,
			   _word_limit, _use_pos);
      }
      if(_use_token_context) {
	fg_token_context::phi(sent_x, tok_feats_str, 2,
			      _word_limit, _use_pos);
      }
      /* map the features to integers */
      for(int i = 0; i < N; ++i) {
	_tok_dict.map_all(tok_feats[i], tok_feats_str[i], _generation_mode);
      }
      _tok_dict.map_all(root_feats, root_feats_str, _generation_mode);
      delete [] tok_feats_str;
    }

    /* dependency features */
    list<string>* dep_feats = new list<string>[NN];
    if(_use_dependency) {
      fg_dependency::phi(sent_x, dep_feats,
			 _word_limit, _use_pos);
    }
    if(_use_dependency_context) {
      fg_dependency_context::phi(sent_x, dep_feats,
				 _word_limit, _use_pos);
    }
    if(_use_dependency_distance) {
      fg_dependency_distance::phi(sent_x, dep_feats,
				  _word_limit, _use_pos);
    }
    if(_use_dependency_between) {
      fg_dependency_between::phi(sent_x, dep_feats,
				 _word_limit, _use_pos);
    }


    /* copy to the feature vectors */
    struct fvec* const ret = (struct fvec*)malloc(NNK*sizeof(struct fvec));
    for(int h = 0; h < N; ++h) {
      for(int m = 0; m < N; ++m) {
	/* there are features for the head-token, mod-token, and the
	   dependency itself */
	const vector<int>& f_h = (h == m ? root_feats : tok_feats[m]);
	const vector<int>& f_m = tok_feats[m];
	const int dep_idx = h + N*m;
	const list<string>& flist_dep = dep_feats[dep_idx];
	vector<int> f_dep;
	_dep_dict.map_all(f_dep, flist_dep, _generation_mode);
	const int nfeats = f_h.size() + f_m.size() + f_dep.size();
	/* allocate the feature array */
	int* const f = (int*)malloc(nfeats*sizeof(int));
	/* copy in the head features */
	vector<int>::const_iterator it = f_h.begin();
	vector<int>::const_iterator it_end = f_h.end();
	int j = 0;
	for(; it != it_end; ++it, ++j) { f[j] = *it; }
	/* copy in the modifier features */
	it = f_m.begin();
	it_end = f_m.end();
	for(; it != it_end; ++it, ++j) { f[j] = *it; }
	/* copy in the dependency features */
	it = f_dep.begin();
	it_end = f_dep.end();
	for(; it != it_end; ++it, ++j) { f[j] = *it; }
	assert(j == nfeats);
	/* assign values to the output feature vectors */
	struct fvec* fv = ret + dep_idx;
	for(int k = 0; k < _K; ++k, fv += NN) {
	  /* we set all feature vectors for a given label to point to
	     the same underlying arrays.  this allows us to avoid
	     having to create several copies of what are essentially
	     the same features.  the offset field allows the different
	     labels to be redirected to different regions of the
	     feature space. */
	  fv->idx = f;
	  fv->val = NULL; /* indicator features */
	  fv->n = nfeats;
	  fv->offset = _unlab_dim*k;
	}
      }
    }

    delete [] tok_feats;
    delete [] dep_feats;

    return ret;
  }

  void depfeats1::discard(const struct fvec* const F,
			  const pattern* const x) {
    const int N = x->size();
    const int NN = N*N;

    /* the underlying vectors are identical for all labels k, so it
       suffices to delete the vectors for just the first label */
    for(int r = 0; r < NN; ++r) { free((void*)(F[r].idx)); }
    free((void*)F);
  }
}
