/* StructuredModel : abstract class representing a part-factorized
   structured model

   Create a subclass of this class and define whichever of the
   inference routines you need.  To classify data, you will need to
   define the argmax() routine.  To use EG, you will need to define
   the marginals() routine.  To train log-linear models in the primal,
   you will need to define the marginals() and partition() routines.

   It is important to have a consistent method for numbering the parts
   in a given pattern.

   You should redefine the name to be descriptive of the subclass, for
   debug-output purposes.

   The usage() and process_options() functions allow the model to
   handle command-line options (see Options.h/cc).

   Author : Terry Koo
            maestro@csail.mit.edu */

#ifndef EGSTRA_STRUCTURED_MODEL_H
#define EGSTRA_STRUCTURED_MODEL_H

#include "Pattern.h"
#include "Label.h"

#include <stdio.h>

namespace egstra {
  class structured_model {
  protected:
    const char* _name;

  public:
    structured_model(const char* name) : _name(name) {}
    virtual ~structured_model() {}

    /* handle the command line */
    virtual void usage(const char* const mesg = "") = 0;
    virtual void process_options() = 0;

    /* accessors */
    const char* name() { return _name; }

    /* read the next (x_i,y_i) pair out of the provided data file.  if
       the are no more pairs, then return false, otherwise set the
       values of x and y and return true. */
    virtual bool fread(FILE* in, pattern*& x, label*& y) = 0;
    /* write the (x,y) pair to the provided file.  this routine will
       be used to save decoded labels for held-out validation data. */
    virtual void fwrite(FILE* out, const pattern* const x, const label& y) = 0;

    virtual double argmax(const pattern* x, const double* const S, label& y);
    virtual void kbest_init(const pattern* x, const double* const S);
    virtual void kbest_clear();
    virtual bool kbest(int k, const pattern* x, const double* const S,
		       label& y, double& score);
    virtual double partition(const pattern* x, const double* const S);
    virtual double marginals(const pattern* x, const double* const S,
			     double* const M);
  };
}

#endif /* EGSTRA_STRUCTURED_MODEL_H */
