#include "ProjDep1.h"

#include "Options.h"
#include "Constants.h"
#include "CharUtils.h"
#include "CppAssert.h"
#include "Sentence.h"

#include <iostream>
#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>


/* The part-numbering for labeled dependencies is defined as
   follows:

   part-index(h,m,l) = h + N*m + N*N*l

   where h is the head word index, m is the modifier word index, l is
   the label index, and N is the length of the sentence (excluding
   root token).  Root dependencies are numbered with h == m. */


using namespace std;

namespace egstra {

  void projdep1::usage(const char* const mesg) {
    cerr << "ProjDep1 options:" << endl;
    cerr << " --multiroot : allow multiple roots in parsing" << endl;
    cerr << " --L=<int>   : number of dependency labels (default 1)" << endl;
    cerr << endl;
    cerr << mesg << endl;
  }

  void projdep1::process_options() {
    int tmp;
    _multiroot = false;
    if(options::get("multiroot", tmp)) {
      _multiroot = true;
    }

    _K = -1;
    if(!options::get("L", _K)) {
      usage("please provide the number of dependency labels"); exit(1);
    }
    assert(_K > 0);
  }


  /***************************************************************
   * The following inference routines were written by Xavier Carreras
   ***************************************************************/


  /* Back pointers to recover the best deptree from DynProg
   * optimization matrix
   *
   * The pointers are indexed by head and modifier
   * Internally, pointers for head=-1 (root) are stored in the [m,m] position
   *
   */
  struct projdep1::backp {
    int _N;   /* number of words */
    int *_RC;  /* optimal splitting points for complete signatures */
    int *_RU;  /* optimal splitting points for uncomplete signatures */
    int *_K;   /* optimal labels for uncomplete signatures */

    backp(int n0) {
      _N = n0;
      _RC = new int[_N*_N];
      _RU = new int[_N*_N];
      _K = new int[_N*_N];
    }

    ~backp() {
      delete [] _RC;
      delete [] _RU;
      delete [] _K;
    }

    void set_r_com(int h, int m, int r) {
      if (h==-1) h=m;

      _RC[m*_N + h] = r;
    }
    int get_r_com(int h, int m) const {
      if (h==-1) h=m;

      return _RC[m*_N + h];
    }
    void set_r_unc(int h, int m, int r) {
      if (h==-1) h=m;

      _RU[m*_N + h] = r;
    }
    int get_r_unc(int h, int m) const {
      if (h==-1) h=m;

      return _RU[m*_N + h];
    }
    void set_k(int h, int m, int k) {
      if (h==-1) h=m;

      _K[m*_N + h] = k;
    }
    int get_k(int h, int m) const {
      if (h==-1) h=m;

      return _K[m*_N + h];
    }
  };


  /* a set of part-wise marginal probabilities, represented in the
     log-domain */
  class projdep1::marg {
  private:

    int _N, _K;

    double *_ICroot, *_OCroot, *_IUroot, *_OUroot;
    double *_IC, *_IU, *_OC, *_OU;

  public:
    marg(int k0, int n0)
      : _N(n0), _K(k0)
    {
      _IC = new double[_N*_N];
      _IU = new double[_N*_N*_K];
      _OC = new double[_N*_N];
      _OU = new double[_N*_N];

      _IUroot = new double[_N*_K];
      _ICroot = new double[_N];
      _OUroot = new double[_N];
      _OCroot = new double[_N];

      int i;

      double l = log(1.0);
      for (i=0; i<_N; i++) {
	_ICroot[i] = _OCroot[i] = _OUroot[i] = l;
      }
      for (i=0; i<_N*_K; i++) {
	_IUroot[i] = l;
      }

      for (i=0; i<_N*_N; i++) {
	_IC[i] = _OC[i] = _OU[i] = l;
      }
      for (i=0; i<_N*_N*_K; i++) {
	_IU[i] = l;
      }
    }

    ~marg() {
      delete [] _IC;
      delete [] _OC;
      delete [] _IU;
      delete [] _OU;

      delete [] _ICroot;
      delete [] _OCroot;
      delete [] _IUroot;
      delete [] _OUroot;
    }


    // i,j range from -1 to n
    static double log_add(double a, double b)
    {
      if (a>b)
	return a+log(1+exp(b-a));
      else
	return b+log(1+exp(a-b));
    }

    int N() const { return _N; }
    int K() const { return _K; }

    void inside_C_add(int h, int m, double w) { _IC[h*_N + m] = log_add(_IC[h*_N + m],w); }
    void outside_C_add(int h, int m, double w) { _OC[h*_N + m] = log_add(_OC[h*_N + m],w); }

    void inside_U_add(int h, int m, int k, double w) { _IU[h*_N*_K + m*_K + k] = log_add(_IU[h*_N*_K + m*_K + k], w); }
    void outside_U_add(int h, int m, double w) { _OU[h*_N + m] = log_add(_OU[h*_N + m], w); }

    void inside_C_set(int h, int m, double w) { _IC[h*_N + m] = w; }
    void outside_C_set(int h, int m, double w) { _OC[h*_N + m] = w; }

    void inside_U_set(int h, int m, int k, double w) { _IU[h*_N*_K + m*_K + k] = w; }
    void outside_U_set(int h, int m, double w) { _OU[h*_N + m] = w; }

    double inside_C(int h, int m) const { return _IC[h*_N + m]; }
    double outside_C(int h, int m) const { return _OC[h*_N + m]; }
    double inside_U(int h, int m, int k) const { return _IU[h*_N*_K + m*_K + k]; }
    double outside_U(int h, int m) const { return _OU[h*_N + m ]; }


    double inside_C_root(int m) const { return _ICroot[m]; }
    double inside_U_root(int i, int k) const { return _IUroot[i*_K + k]; }
    double outside_C_root(int m) const { return _OCroot[m]; }
    double outside_U_root(int i) const { return _OUroot[i]; }

    void inside_C_root_set(int m, double w) { _ICroot[m] = w; }
    void inside_C_root_add(int m, double w) { _ICroot[m] = log_add(_ICroot[m], w); }
    void inside_U_root_set(int m, int k, double w) { _IUroot[m*_K +k] = w; }
    void outside_C_root_set(int m, double w) { _OCroot[m] = w; }
    void outside_C_root_add(int m, double w) { _OCroot[m] = log_add(_OCroot[m], w); }
    void outside_U_root_set(int m, double w) { _OUroot[m] = w; }
    void outside_U_root_add(int m, double w) { _OUroot[m] = log_add(_OUroot[m], w); }

    double log_Z() const { return _ICroot[_N-1]; }

    double log_marginal(int h, int m, int l) const {
      if (h==-1)
	return inside_U_root(m,l) + outside_U_root(m) - _ICroot[_N-1];
      else
	return inside_U(h,m,l) + outside_U(h,m) - _ICroot[_N-1];
    }


    void write(ostream & out, bool E) const {
      int o,i,j,k;

      // INSIDES

      if (E)
	out << "I C [" << -1 << "," << _N-1 << "] :: " << exp(this->inside_C_root(_N-1)) << endl;
      else
	out << "I log C [" << -1 << "," << _N-1 << "] :: " << this->inside_C_root(_N-1) << endl;

      for (i=0; i<_N; ++i) {
	if (E) {
	  out << "I [" << -1 << "," << i << "] ::";
	  for (k=0; k<_K; ++k) {
	    out << "  (ul_" << k << ") " << exp(this->inside_U_root(i,k));
	  }
	  out << endl;
	}
	else {
	  out << "I log [" << -1 << "," << i << "] ::";
	  for (k=0; k<_K; ++k) {
	    out << "  (ul_" << k << ") " << this->inside_U_root(i,k);
	  }
	  out << endl;
	}
      }

      for (o=1; o<_N; ++o) {
	for (i=0; i<_N-o; ++i) {
	  j = i + o;
	  if (E) {
	    out << "I [" << i << "," << j << "] ::";
	    out << "  (cl) " << exp(this->inside_C(i,j)) << "  (cr) " << exp(this->inside_C(j,i));
	    for (k=0; k<_K; ++k) {
	      out << "  (ul_" << k << ") " << exp(this->inside_U(i,j,k)) << "  (ur_" << k << ") " << exp(this->inside_U(j,i,k));
	    }
	    out << endl;
	  }
	  else {
	    out << "I log [" << i << "," << j << "] ::";
	    out << "  (cl) " << this->inside_C(i,j) << "  (cr) " << this->inside_C(j,i);
	    for (k=0; k<_K; ++k) {
	      out << "  (ul_" << k << ") " << this->inside_U(i,j,k) << "  (ur_" << k << ") " << this->inside_U(j,i,k);
	    }
	    out << endl;
	  }
	}
      }

      // OUTSIDES

      if (E)
	out << "O C [" << -1 << "," << _N-1 << "] :: " << exp(this->outside_C_root(_N-1)) << endl;
      else
	out << "O log C [" << -1 << "," << _N-1 << "] :: " << this->outside_C_root(_N-1) << endl;

      for (i=0; i<_N; ++i) {
	if (E) {
	  out << "O [" << -1 << "," << i << "] ::";
	  out << "  (cl) " << exp(this->outside_U_root(i)) << endl;
	}
	else {
	  out << "O log [" << -1 << "," << i << "] ::";
	  out << "  (cl) " << this->outside_U_root(i) << endl;
	}
      }

      for (o=1; o<_N; ++o) {
	for (i=0; i<_N-o; ++i) {
	  j = i + o;
	  if (E) {
	    out << "O [" << i << "," << j << "] ::";
	    out << "  (cl) " << exp(this->outside_C(i,j)) << "  (cr) " << exp(this->outside_C(j,i));
	    out << "  (ul) " << exp(this->outside_U(i,j)) << "  (ur) " << exp(this->outside_U(j,i)) << endl;
	  }
	  else {
	    out << "O log [" << i << "," << j << "] ::";
	    out << "  (cl) " << this->outside_C(i,j) << "  (cr) " << this->outside_C(j,i);
	    out << "  (ul) " << this->outside_U(i,j) << "  (ur) " << this->outside_U(j,i) << endl;
	  }
	}
      }

    }

  }; /* end marg class definition */


  void projdep1::unravel_tree(label& y, const backp* const BP,
			      const int h, const int m,
			      const bool complete, const int N) {
    //    cout << "utree : h " << h << " m " << m << " c? " << complete << flush;
    if (h==m) {
      //      cout << endl;
      return;
    }

    if (complete) {
      int r = BP->get_r_com(h,m);
      //      cout << " ; complete r " << r << endl;
      unravel_tree(y, BP, h, r, false, N);
      unravel_tree(y, BP, r, m, true, N);
    }
    else {
      int r = BP->get_r_unc(h,m);
      int k = BP->get_k(h,m);
      //      cout << " ; uncomplete r " << r << " k " << k << endl;
      y.push_back((h == -1 ? m : h) + N*m + N*N*k);
      /*dtree.set_dependency(h, m, k);*/
      if (h<m) {
	unravel_tree(y, BP, h, r, true, N);
	unravel_tree(y, BP, m, r+1, true, N);
      }
      else {
	unravel_tree(y, BP, m, r, true, N);
	unravel_tree(y, BP, h, r+1, true, N);
      }
    }
  }



  double projdep1::argmax(const pattern* x,
			  const double* const S,
			  label& y){
    const int N = x->size();
    const int NN = N*N;
    const int NNK = NN*_K;
    assert(NNK == x->nparts());

    // matrices of scores
    // SCom : for complete structures
    // SUnc : for uncomplete structures
    // 1st/2nd dimension : head/modifier positions
    double** SCom = new double*[N];
    double** SUnc = new double*[N];
    for(int i = 0; i < N; ++i) {
      SCom[i] = new double[N];
      SUnc[i] = new double[N];
    }

    double* SRCom = new double[N];
    double* SRUnc = new double[N];

    // a structure of back-pointers to recover the viterbi tree
    backp BP(N);

    int s,e,w;

    // initialization
    for (s=0;s<N;++s) {
      SCom[s][s] = 0.0;
      SUnc[s][s] = 0.0;
    }

    /*
     * W is the max width of the spans it controls the outer loop of
     * the viterbi search,
     *
     * when multiple roots are allowed, its value is N+1, so that
     * dependencies from the root are visited from token 0 to token
     * N-1
     *
     */
    int W = N;
    if (_multiroot) W++;

    for (w=1; w<W; ++w) {

      /*
       * The code in this conditional treats the structures headed at the root.
       * The operations are the same than with regular tokens, except for two
       * issues: the root has its own dyn_prog matrix, and the root only can be
       * head, not modifier
       */
      if (_multiroot) {
	e = -1+w;

	int r;                         /* iterator for splitting points between -1 and e */
	int opt_r = -1;                /* optimal splitting point, initialized to -1 */
	double opt_sco = SCom[e][0];   /* score the optimal splitting point */

	// search for best incomplete dtree
	for (r=0; r<e; ++r) {
	  if (SRCom[r] + SCom[e][r+1] > opt_sco) {
	    opt_r = r;
	    opt_sco = SRCom[r] + SCom[e][r+1];
	  }
	}
	int opt_k = 0;
	double opt_ksco = S[e + N*e];/*scores.score(-1,e,0);*/
	for (int k=1; k<_K; ++k) {
	  const double sc = S[e + N*e + NN*k];/*scores.score(-1,e,k)*/
	  if (sc > opt_ksco) {
	    opt_k = k;
	    opt_ksco = sc;
	  }
	}
	SRUnc[e] = opt_sco + opt_ksco;
	BP.set_r_unc(-1,e,opt_r);
	BP.set_k(-1,e,opt_k);

	// complete structure
	opt_r = -1;
	for (r=0; r<=e; ++r) {
	  if ((opt_r==-1) || (SRUnc[r] + SCom[r][e] > opt_sco)) {
	    opt_r = r;
	    opt_sco = SRUnc[r] + SCom[r][e];
	  }
	}
	SRCom[e] = opt_sco;
	BP.set_r_com(-1,e,opt_r);
      }

      for (s=0; s<N-w; ++s) {
	e = s+w;

	int r;                           /* iterator for splitting points between s and e */
	int opt_r = s;                   /* optimal splitting point, initialized to r=s */
	double opt_sco = SCom[e][s+1];   /* score the optimal splitting point */

	// search for best incomplete dtree
	for (r=s+1; r<e; ++r) {
	  if (SCom[s][r] + SCom[e][r+1] > opt_sco) {
	    opt_r = r;
	    opt_sco = SCom[s][r] + SCom[e][r+1];
	  }
	}


	// Dependency headed at s

	int k;                  /* iterator for classes */
	int opt_k;              /* optimal class */
	double opt_ksco = 0;    /* score of the optimal class */

	opt_k = 0;
	opt_ksco = S[s + N*e];/*scores.score(s,e,0);*/
	for (k=1; k<_K; ++k) {
	  const double sc = S[s + N*e + NN*k];/*scores.score(s,e,k)*/
	  if (sc > opt_ksco) {
	    opt_k = k;
	    opt_ksco = sc;
	  }
	}
	SUnc[s][e] = opt_sco + opt_ksco;
	BP.set_r_unc(s,e,opt_r);
	BP.set_k(s,e,opt_k);


	// Dependency headed at e

	opt_k = 0;
	opt_ksco = S[e + N*s];/*scores.score(e,s,0);*/
	for (k=1; k<_K; ++k) {
	  const double sc = S[e + N*s + NN*k];/*scores.score(e,s,k)*/
	  if (sc>opt_ksco) {
	    opt_k = k;
	    opt_ksco = sc;
	  }
	}
	SUnc[e][s] = opt_sco + opt_ksco;
	BP.set_r_unc(e,s,opt_r);
	BP.set_k(e,s,opt_k);


	// search for best complete dtree with head at s
	opt_r = -1;
	for (r=s+1; r<=e; ++r) {
	  if ((opt_r==-1) || (SUnc[s][r] + SCom[r][e] > opt_sco)) {
	    opt_r = r;
	    opt_sco = SUnc[s][r] + SCom[r][e];
	  }
	}
	SCom[s][e] = opt_sco;
	BP.set_r_com(s,e,opt_r);

	// search for best complete dtree with head at e
	opt_r = -1;
	for (r=s; r<e; ++r) {
	  if ((opt_r==-1) || (SCom[r][s] + SUnc[e][r] > opt_sco)) {
	    opt_r = r;
	    opt_sco = SCom[r][s] + SUnc[e][r];
	  }
	}
	SCom[e][s] = opt_sco;
	BP.set_r_com(e,s,opt_r);
      }
    }

    double global_score = 0;


    // Root
    if (_multiroot) {
      /* here, the work has already been done before */
      global_score = SRCom[N-1];
      bool complete = true;
      unravel_tree(y, &BP, -1, N-1, complete, N);
    }
    else {
      /* consider the dependencies that go from the root to any of the
	 tokens, and choose the optimal */
      int opt_root = -1;
      int opt_rootK = -1;

      for (s=0; s<N; ++s) {
	int k;                /* iterator for classes */
	int opt_k = 0;        /* optimal class */
	double opt_sco = S[s + N*s];/*scores.score(-1,s,0);*/ /* score of the optimal class */
	for (k=1; k<_K; ++k) {
	  const double sc = S[s + N*s + NN*k];/*scores.score(-1,s,k)*/
	  if (sc > opt_sco) {
	    opt_k = k;
	    opt_sco = sc;
	  }
	}

	opt_sco += SCom[s][0] + SCom[s][N-1];

	if (opt_root==-1 or global_score < opt_sco) {
	  opt_root = s;
	  opt_rootK = opt_k;
	  global_score = opt_sco;
	}
      }

      //      cout << "P sentence parsed : root is " << opt_root << "; score is " << global_score << endl;

      /*dtree.set_dependency(-1, opt_root, opt_rootK);*/
      y.push_back(opt_root + N*opt_root + NN*opt_rootK);
      unravel_tree(y, &BP, opt_root, 0, true, N);
      unravel_tree(y, &BP, opt_root, N-1, true, N);
    }

    for(int i = 0; i < N; ++i) {
      delete [] SCom[i];
      delete [] SUnc[i];
    }
    delete [] SCom;
    delete [] SUnc;

    delete [] SRCom;
    delete [] SRUnc;

    return global_score;
  }





  void projdep1::outside(const double* const S, marg& M){

    const int N = M.N();
    const int NN = N*N;

    int w,s,e,k,r;
    double aux, auxL, auxR;


    // Root
    M.outside_C_root_set(N-1, (double) log(1.0));

    if (!_multiroot) {
      for (s=0; s<N; ++s) {
	M.outside_U_root_set(s, M.outside_C_root(N-1) + M.inside_C(s,N-1));
      }
    }

    int W = _multiroot ? N : N-1;
    for (w=W; w>0; --w) {

      /* Structures headed at the root token, treated specially when multiple roots are possible*/
      if (_multiroot) {
	e = -1+w;

	// complete signature
	int c = 0;   /* initialized? */
	for (r=e+1; r<N; ++r) {
	  aux = M.outside_U_root(r) + M.inside_C(r,e+1);
	  for (k=0; k<_K; ++k) {
	    if (!c++)
	      M.outside_C_root_set(e, aux + S[r + N*r + NN*k]);
	    else
	      M.outside_C_root_add(e, aux + S[r + N*r + NN*k]);
	  }
	}

	// uncomplete
	/* initialize with r=e */
	M.outside_U_root_set(e, M.outside_C_root(e) + M.inside_C(e,e));
	for (r=e+1; r<N; ++r) {
	  M.outside_U_root_add(e, M.outside_C_root(r) + M.inside_C(e,r) );
	}
      }

      /* Structures spanning from s to e */
      for (s=0; s<N-w; ++s) {
	e = s+w;

	/// COMPLETE signatures

	// when c==0 the cell is initialized (with set), otherwise new values are added to the existing ones
	int c = 0;

	// LEFT
	// first the structures coming from the root;
	if (_multiroot || e==N-1) {
	  // initialization with k=0
	  c = 1;
	  M.outside_C_set(s,e, M.outside_C_root(e) + M.inside_U_root(s,0));
	  for (k=1; k<_K; ++k) {
	    M.outside_C_add(s,e, M.outside_C_root(e) + M.inside_U_root(s,k));
	  }
	}

	for (r=0; r<s; ++r) {
	  for (k=0; k<_K; ++k) {
	    if (!c++)
	      M.outside_C_set(s,e, M.outside_C(r,e) + M.inside_U(r,s,k) );
	    else
	      M.outside_C_add(s,e, M.outside_C(r,e) + M.inside_U(r,s,k) );
	  }
	}

	for (r=e+1; r<N; ++r) {
	  auxL = M.outside_U(s,r) + M.inside_C(r,e+1);

	  for (k=0; k<_K; ++k) {
	    if (!c++)
	      M.outside_C_set(s,e, auxL + S[s + N*r + NN*k]);
	    else
	      M.outside_C_add(s,e, auxL + S[s + N*r + NN*k]);
	  }

	  auxR = M.outside_U(r,s) + M.inside_C(r,e+1);
	  for (k=0; k<_K; ++k) {
	    M.outside_C_add(s,e, auxR + S[r + N*s + NN*k]);
	  }
	}


	// RIGHT
	c = 0;
	// first the structures coming from the root;
	if (s==0 || _multiroot) {
	  // initialization with k=0
	  c = 1;
	  aux = M.outside_U_root(e);
	  if (s>0) aux += M.inside_C_root(s-1);
	  M.outside_C_set(e,s, aux + S[e + N*e]);
	  for (k=1; k<_K; ++k) {
	    M.outside_C_add(e,s, aux +  S[e + N*e + NN*k]);
	  }
	}

	for (r=e+1; r<N; ++r) {
	  for (k=0; k<_K; ++k) {
	    if (!c++)
	      M.outside_C_set(e,s, M.outside_C(r,s) + M.inside_U(r,e,k) );
	    else
	      M.outside_C_add(e,s, M.outside_C(r,s) + M.inside_U(r,e,k) );
	  }
	}

	for (r=0; r<s; ++r) {
	  auxL = M.outside_U(r,e) + M.inside_C(r,s-1);
	  for (k=0; k<_K; ++k) {
	    if (!c++)
	      M.outside_C_set(e,s, auxL + S[r + N*e + NN*k]);
	    else
	      M.outside_C_add(e,s, auxL + S[r + N*e + NN*k]);
	  }

	  auxR = M.outside_U(e,r) + M.inside_C(r,s-1);
	  for (k=0; k<_K; ++k) {
	    M.outside_C_add(e,s, auxR + S[e + N*r + NN*k]);
	  }

	}


	/// uncomplete signatures

	// LEFT part : initialize with r=e
	M.outside_U_set(s,e, M.outside_C(s,e) + M.inside_C(e,e));

	// rest of splitting points
	for (r=e+1; r<N; ++r) {
	  M.outside_U_add(s,e, M.outside_C(s,r) + M.inside_C(e,r));
	}

	// RIGHT part : initialize with r=s
	M.outside_U_set(e,s, M.outside_C(e,s) + M.inside_C(s,s));

	// rest of splitting points
	for (r=0; r<s; ++r) {
	  M.outside_U_add(e,s, M.outside_C(e,r) + M.inside_C(s,r));
	}
      }

    }


  }


  void projdep1::inside(const double* const S, marg& M){

    const int N = M.N();
    const int NN = N*N;
    int w,s,e,k,r;


    int W = N;
    if (_multiroot) W++;

    for (w=1; w<W; ++w) {

      /*
       * The code in this conditional treats the structures headed at the root.
       * The operations are the same than with regular tokens, except for two
       * issues: the root has its own inside calls, and the root only can be
       * head, not modifier
       */
      if (_multiroot) {
	e = -1+w;

	double psco = M.inside_C(e,0);   /*  */

	for (r=0; r<e; ++r) {
	  psco = marg::log_add(psco, M.inside_C_root(r) + M.inside_C(e,r+1));
	}

	// uncomplete structures
	for (k=0; k<_K; ++k) {
	  M.inside_U_root_set(e, k, psco + S[e + N*e + NN*k]);
	}

	// complete structure
	for(r=0; r<=e; ++r) {
	  double cl = M.inside_C(r,e);
	  for (k=0; k<_K; ++k) {
	    if (r==0 && k==0)
	      M.inside_C_root_set(e, M.inside_U_root(r,k) + cl);
	    else
	      M.inside_C_root_add(e, M.inside_U_root(r,k) + cl);
	  }
	}
      }

      for (s=0; s<N-w; ++s) {
	e = s+w;

	double psco = M.inside_C(s,s) + M.inside_C(e,s+1);
	for (r=s+1; r<e; ++r) {
	  psco = marg::log_add(psco, M.inside_C(s,r) + M.inside_C(e,r+1));
	}

	// headed at s
	for (k=0; k<_K; ++k) {
	  M.inside_U_set(s,e,k, psco + S[s + N*e + NN*k]);
	}

	// headed at e
	if (s!=-1) {
	  for (k=0; k<_K; ++k) {
	    M.inside_U_set(e,s,k, psco + S[e + N*s + NN*k]);
	  }
	}

	// Complete spans

	for(r=s+1; r<=e; ++r) {
	  double cl = M.inside_C(r,e);
	  for (k=0; k<_K; ++k) {
	    if (r==s+1 && k==0)
	      M.inside_C_set(s,e, M.inside_U(s,r,k) + cl);
	    else
	      M.inside_C_add(s,e, M.inside_U(s,r,k) + cl);
	  }
	  //	  cout << "INSIDE C Left " << s << " " << e << " " << r << " .... " << M.inside_C(s,e) << endl;
	}
	//	cout << "INSIDE C Left " << s << " " << e << " is " << M.inside_C(s,e) << endl;

	for(r=s; r<e; ++r) {
	  double cr = M.inside_C(r,s);
	  for (k=0; k<_K; ++k) {
	    if (r==s && k==0)
	      M.inside_C_set(e,s, cr + M.inside_U(e,r,k));
	    else
	      M.inside_C_add(e,s, cr + M.inside_U(e,r,k));
	  }
	}

      }
    }

    if (!_multiroot) {
      // STRUCTURES AT ROOT WORD
      for(s=0; s<N; ++s) {
	for (k=0; k<_K; ++k) {
	  double aux = S[s + N*s + NN*k] + M.inside_C(s,0);
	  M.inside_U_root_set(s, k, aux );

	  if (s==0 && k==0)
	    M.inside_C_root_set(N-1, aux + M.inside_C(s,N-1) );
	  else
	    M.inside_C_root_add(N-1, aux + M.inside_C(s,N-1) );
	}
      }
    }

  }




  double projdep1::partition(const pattern* x, const double* const S) {
    const int N = x->size();
    const int NNK = N*N*_K;
    assert(NNK == x->nparts());
    marg M(_K, N);
    inside(S, M);
    return M.log_Z();
  }



  double projdep1::marginals(const pattern* x, const double* const S,
			     double* const out_M) {
    const int N = x->size();
    const int NN = N*N;
    const int NNK = NN*_K;
    assert(NNK == x->nparts());
    marg M(_K, N);
    inside(S, M);
    outside(S, M);
    /* copy the log-marginals to the output array */
    for(int k = 0; k < _K; ++k) {
      double* const M_k = out_M + NN*k;
      for(int m = 0; m < N; ++m) {
	double* const M_mk = M_k + N*m;
	for(int h = 0; h < N; ++h) {
	  const int hh = (h == m ? -1 : h);
	  M_mk[h] = exp(M.log_marginal(hh, m, k));
	}
      }
    }
    return M.log_Z();
  }


  /**************************************************************
   *                end of inference routines
   **************************************************************/

  bool projdep1::fread(FILE* in, pattern*& ret_x, label*& ret_y) {
    sentence* x = new sentence;
    label* y = new label;

    char line[MAX_CONLLX_LINE];
    vector<int> heads;
    vector<int> labels;
    int cnt = 0;
    while(1) {
      fgets(line, MAX_CONLLX_LINE, in);
      const int len = strlen(line);
      if(len == 0) { break; }
      { /* check for whitespace-only lines */
	int i;
	for(i = 0; i < len && (line[i] == ' ' || line[i] == '\t'); ++i) {}
	if(line[i] == '\n') { break; }
      }
      if(line[len - 1] != '\n') {
	cerr << "detected line overflow when reading data" << endl;
	assert(0);
      }

      /* split along whitespace and create a token */
      vector<string> tokens;
      simpleTokenize(line, tokens, " \t");
      RVASSERT(tokens.size() >= 8, "Invalid corpus line: " << line);
      token t(tokens[1], tokens[2], tokens[3],
	      tokens[4], tokens[5]);
      x->add_token(t);

      /* read the parent dependency */
      const int h = toInteger(tokens[6]) - 1;
      RVASSERT(h >= -1, "Invalid head position: " << line);
      const int l = toInteger(tokens[7]);
      RVASSERT(l >= 0, "Invalid label index: " << line);
      heads.push_back(h);
      labels.push_back(l);
      ++cnt;
    }

    /* empty sentence: no more sentences left */
    if(cnt == 0) {
      delete x;
      delete y;
      return false;
    }

    /* set size variables */
    const int N = cnt;
    x->set_size(N);
    const int NN = N*N;
    x->set_nparts(NN*_K);

    /* convert the (head,label) pairs to part indices */
    for(int m = 0; m < N; ++m) {
      /* h=-1 is the root token; translate to h=m */
      const int h = (heads[m] != -1 ? heads[m] : m);
      /* k=1 is unlabeled; translate to l=0 */
      const int l = (_K == 1 ? 0 : labels[m]);
      assert(h < N);
      assert(h >= 0);
      assert(l < _K);
      assert(l >= 0);
      const int r = h + N*m + NN*l;
      y->push_back(r);
    }

    ret_x = x;
    ret_y = y;
    return true;
  }

  void projdep1::fwrite(FILE* out, const pattern* const x, const label& y) {
    const int N = x->size();
    const int NN = N*N;
    const int NNK = NN*_K;
    assert(NNK == x->nparts());

    /* create the parent-array representation */
    int* parent = new int[N];
    int* label  = new int[N];
    for(int i = 0; i < N; ++i) { parent[i] = label[i] = -2; }
    label::const_iterator it = y.begin();
    const label::const_iterator it_end = y.end();
    for(; it != it_end; ++it) {
      const int r = *it;
      assert(r >= 0);
      assert(r < NNK);
      /* r = h + N*m */
      const int h = r % N;
      const int m = (r % NN)/N;
      const int k = r/NN;
      assert(h >= 0);
      assert(h < N);
      assert(m >= 0);
      assert(m < N);
      assert(k >= 0);
      assert(k < _K);
      parent[m] = (h == m ? -1 : h);
      label[m] = k;
    }
    /* check that all parents are set */
    for(int i = 0; i < N; ++i) {
      assert(parent[i] != -2);
      assert(label[i] != -2);
    }

    /* print the parent array */
    for(int i = 0; i < N; ++i) {
      fprintf(out, "%d %d %d %d\n",
	      x->id(), i + 1, parent[i] + 1, label[i]);
    }
    fprintf(out, "\n");

    delete [] parent;
    delete [] label;
  }
}
