#include "Timer.h"
#include "OptUtils.h"
#include "Options.h"
#include "RegPrimal.h"
#include "Constants.h"

#include <iostream>
#include <iomanip>
#include <fstream>

//#define REPORT_EACH_PROCESSED_EXAMPLE

/* detailed name of the trainer */
#define myname								\
  (_adapt > 0 ? "A" : "")						\
  << RAND_SHORT_GLOSS(_rand_type)					\
  << "DEG[" << _M->name() << "," << Obj::name << "]"

using namespace std;

namespace egstra {
  template <class Obj>
  void deg_trainer<Obj>::usage(const char* const mesg) {
    cerr << "DEGTrainer[" << Obj::name << "] options:" << endl;
    cerr << " --rand=<string> : randomization type (det,unif,perm)" << endl;
    cerr << " --seed=<int>    : random seed" << endl;
    cerr << " --C=<real>      : regularization constant" << endl;
    cerr << " --E=<real>      : initial learning rate.  if not provided or" << endl;
    cerr << "                   if set to a negative value, then the initial" << endl;
    cerr << "                   rate is estimated, which generally works" << endl;
    cerr << "                   well.  the learning rate is fixed by default" << endl;
    cerr << "                   (unless adapting)" << endl;
    cerr << " --adapt=<real>  : use a per-example adaptive learning rate" << endl;
    cerr << "                   which multiplies the learning rate by the" << endl;
    cerr << "                   given value after each successful update" << endl;
    cerr << " --adaptive      : synonym for adapt="
	 << EGSTRA_LRATE_DEFAULT_BOOST << endl;
    cerr << " --beta=<real>   : bias on correct parts (defaults to "
	 << EGSTRA_BETA_DEFAULT << ")" << endl;
    cerr << " --noparams      : do not write weight vectors to disk" << endl;
    cerr << " --noprimal      : do not compute primal objective" << endl;
  //  cerr << " --anneal-C (perform annealing on regularizer)" << endl;
  //  cerr << " --regpath (compute a regularization path)" << endl;
  //  cerr << endl;

    cerr << endl;
    cerr << mesg << endl;
    exit(1);
  }

  template <class Obj>
  void deg_trainer<Obj>::process_options() {
    int tmpi;
    _seed = 1;
    if(options::get("seed", tmpi)) {
      _seed = tmpi;
    }

    _rand_type = EGSTRA_RAND_UNIFORM;
    string rtype;
    if(options::get("rand", rtype)) {
      if(strcmp(rtype.c_str(), "det") == 0) {
	_rand_type = EGSTRA_DETERMINISTIC;
      } else if(strcmp(rtype.c_str(), "unif") == 0) {
	_rand_type = EGSTRA_RAND_UNIFORM;
      } else if(strcmp(rtype.c_str(), "perm") == 0) {
	_rand_type = EGSTRA_RAND_PERMUTATION;
      } else {
	usage("unrecognized randomization type"); exit(1);
      }
    }

    if(!options::get("C", _C)) {
      usage("please provide a regularization constant"); exit(1);
    }
    if(_C <= 0) {
      usage("C: please provide a value > 0"); exit(1);
    }

    _lrate = -1; /* special meaning: estimate rate */
    double tmpd;
    if(options::get("E", tmpd)) {
      _lrate = tmpd;
      if(_lrate <= 0) {
	usage("E: please provide a value > 0"); exit(1);
      }
    }

    _adapt = -1; /* special meaning: no adaptivity */
    if(options::get("adapt", tmpd)) {
      _adapt = tmpd;
      if(_adapt <= 1) {
	usage("adapt: please provide a value > 1"); exit(1);
      }
    }
    if(options::get("adaptive", tmpd)) {
      _adapt = EGSTRA_LRATE_DEFAULT_BOOST;
    }

    _beta = EGSTRA_BETA_DEFAULT;
    if(options::get("beta", tmpd)) {
      _beta = tmpd;
    }

    _writeparams = true;
    if(options::get("noparams", tmpi)) {
      _writeparams = false;
    }

    _doprimal = true;
    if(options::get("noprimal", tmpi)) {
      _doprimal = false;
    }
  }


  template <class Obj>
  deg_trainer<Obj>::~deg_trainer() {
    if(_duals != NULL)   { delete _duals; }

    if(_lrates != NULL)  { delete [] _lrates; }
    if(_exduals != NULL) { delete [] _exduals; }
  }


  template <class Obj>
  void deg_trainer<Obj>::train(dataset& trdata, dataset& valdata,
			       const int T) {
    _W.realloc(_fg->dim());

    _N = trdata.size();

    cerr << myname << " : training started for max " << T
	 << " rounds ... " << endl;
    cerr << myname << " : data order is "
	 << RAND_LONG_GLOSS(_rand_type) << endl;

    _invNC = 1.0/(_N*_C);

    _cpu_time = 0;

    /* initialize per-example dual value array */
    assert(_exduals == NULL);
    _exduals = new double[_N];
    for(int i = 0; i < _N; ++i) { _exduals[i] = 0; }

    if(_lrate < 0) {
      /* negative learning rate indicates automatic selection */
      const clock_t t0 = clock();
      _lrate = estimate_lrate(trdata);
      _cpu_time += ((double)(clock() - t0))/(double)CLOCKS_PER_SEC;
    } else {
      cerr << myname << " : using provided learning rate " << _lrate << endl;
    }

    /* (re)initialize these values (the learning-rate search might
       have modified them) */
    assert(_duals == NULL);
    _duals = new egduals(_N, false);
    _W.zero();
    for(int i = 0; i < _N; ++i) { _exduals[i] = 0; }

    /* optionally initialize the per-example learning rates */
    assert(_lrates == NULL);
    if(_adapt > 0) {
      cerr << myname << " : per-example learning rate with boost "
	   << setprecision(5) << _adapt << endl;
      _lrates = new double[_N];
      for(int i = 0; i < _N; ++i) { _lrates[i] = _lrate; }
    } else {
      cerr << myname << " : using fixed global learning rate" << endl;
    }


    /* log file for primal/dual and #pass information */
    string objfilename = _dir + "/obj.out";
    ofstream obj_out(objfilename.c_str(), ios::out | ios::trunc);
    assert(obj_out.good());

    if(_rand_type != EGSTRA_DETERMINISTIC) {
      /* only report this for randomized EG */
      cerr << myname << " : seeding the RNG with value " << _seed << endl;
      srandom(_seed);
    }

    _num_processed = 0;

    for(int t = 1; t <= T; ++t) {
      const clock_t t0 = clock();
      const int cur_num_processed = _num_processed;
/* macro that prints #processed/#cur_sent/#total_sent and
   percentages, for use during training */
#define print_stats(ii, nn)						\
	  cerr << "\rt " << t << " : "					\
	       << (_num_processed - cur_num_processed)			\
               << "/" << (ii) << "/" << (nn) << " : "			\
	       << fixed << setprecision(2)				\
	       << 100.0*(double)(_num_processed - cur_num_processed)/(ii) \
               << "/" << 100.0*(double)(ii)/(nn) << flush

      timer_start();
      for(int i = 0; i < _N; ++i) {
	/* first, pick an example according to the desired scheme NB:
	   the first round of rand-uniform will be a permutation, to
	   ensure that all examples are visited. */
	const int rt = (t == 1 && _rand_type == EGSTRA_RAND_UNIFORM
			? EGSTRA_RAND_PERMUTATION : _rand_type);
	const example* const ex = optutils::randselect(rt, i, trdata);
	const pattern* const x = ex->x();

	const struct fvec* const F = _fg->phi(x);

	/* the first pass is the initialization pass (even when
	   rand=unif; see above) */
	if(t == 1) { initialize_example(ex, F); }
	/* perform a dual EG update with line-search */
	process_example(ex, F, t);

	_fg->discard(F, x);

	/* print progress messages */
	if((i & 0xf) == 0) { print_stats(i + 1, _N); }
      }
      print_stats(_N, _N);
#undef print_stats
      timer_stop();
      cerr << " [" << timestr(timediff) << "]" << endl;

      /* NB: after some pilot studies, it seems that the primals do
	 accumulate some error, but it is not necessary to reconstruct
	 the primals, or at least not often.  we only reconstruct
	 every 25 iterations */
      if((t % 25) == 0) { reconstruct_primals(trdata); }

      /* the remainder is not timed */
      _cpu_time += ((double)(clock() - t0))/(double)CLOCKS_PER_SEC;

      /* save the primals in the model directory */
      if(_writeparams) { _W.save(_dir, t); }

      /* get the primal and dual objectives */
      double dual = 0;
      for(int i = 0; i < _N; ++i) { dual += _exduals[i]; }
      dual = dual/_N + 0.5*_C*_W.sqL2(true);
      const double primal =
	(_doprimal
	 ? reg_primal<Obj>::objective(_W, _C, trdata, *_fg, *_M)
	 : 0.0);

      /* parse the validation file if provided */
      if(valdata.size() > 0) {
	optutils::classify_val_set(_W, valdata, *_fg, *_M, _dir, t,
				   Obj::probabilistic());
      }

      /* #passes is an abstract measure of computation */
      const double num_passes = ((double)_num_processed)/((double)_N);
      /* update the objective log with the current iteration */
      obj_out << scientific << setprecision(16)
	      << t << " " << num_passes << " "
	      << _cpu_time << " " << _W.sqL2() << " "
	      << dual << " " << primal << endl;
      /* also report values to stderr (in short format) */
      const double duality_gap = dual - primal;
      cerr << myname << " : finished it = " << t
	   << scientific << setprecision(5)
	   << " ; ops = " << num_passes
	   << " ; cpu = " << _cpu_time
	   << " ; W^2 = " << _W.sqL2()
	   << " ; D = " << dual
	   << " ; P = " << primal
	   << " ; gap = " << duality_gap
	   << endl;

      /* stop the optimization if *computation* exceeds T */
      if(num_passes >= T) {
	cerr << myname << " : number of passes exceeds T=" << T
	     << "; stopping..." << endl;
	break;
      }

      /* stop the optimization if we satisfy the tolerance (in the
	 case that we are annealing the regularization constant, we
	 must be careful to not stop early) */
      const double relative_gap = -duality_gap/primal;
      if(relative_gap <= EGSTRA_CONV_TOL) {
	cerr << myname << scientific << setprecision(5)
	     << " : convergence tolerance satisfied: gap " << relative_gap
	     << " <= " << EGSTRA_CONV_TOL << endl;
	break;
      }
    }

    /* tidy up */
    obj_out.close();
    delete _duals; _duals = NULL;
    delete [] _exduals; _exduals = NULL;
    if(_adapt > 0) {
      delete [] _lrates; _lrates = NULL;
    }
  }


  template <class Obj>
  void deg_trainer<Obj>::initialize_example(const example* const ex,
					    const struct fvec* const F) {
    const pattern* const x = ex->x();
    const int xid = x->id();
    const int R = x->nparts();
    /* create an array of indicators for the correct parts */
    char* const correct = optutils::get_correct(ex);
    /* create the initial theta with the beta bias */
    double* S = new double[R];
    for(int r = 0; r < R; ++r) { S[r] = (correct[r] ? _beta : 0.0); }
    /* compute the part marginals */
    double* P = new double[R];
    const double logZ = _M->marginals(x, S, P);
    /* compute the dual objective and adjust the primals */
    double exdual = Obj::exdual_init(logZ);
    for(int r = 0; r < R; ++r) {
      const int loss = 1 - correct[r];
      exdual += Obj::exdual_contrib(P[r], S[r], loss);
      /* accumulate updates to primals */
      const double contrib = _invNC*(correct[r] - P[r]);
      _W.add(F + r, contrib);
    }
    free(correct);
    /* save the duals */
    _duals->set(xid, S, P, logZ);
    _exduals[xid] = exdual;
  }


  template <class Obj>
  void deg_trainer<Obj>::process_example(const example* const ex,
					 const struct fvec* const F,
					 const int t) {
    const pattern* x = ex->x();
    const int xid = x->id();
    const int R = x->nparts();

    /* create an array of indicators for the incorrect parts */
    char* const loss = optutils::get_correct(ex);
    /* NB: get_correct() returns 0/1 indicators of correct parts, so
       they should be inverted */
    for(int r = 0; r < R; ++r) { loss[r] = 1 - loss[r]; }

    /* get the inner products */
    double* const S = (double*)malloc(R*sizeof(double));
    _W.dot(S, F, R);

    /* get the current duals for this example */
    const double* theta = NULL;
    const double* mu = NULL;
    double logZ = 0;
    _duals->get(xid, theta, mu, logZ);
    const double exdual_cur = _exduals[xid];

    /* allocate space for the new duals for this example */
    double* const theta_prime = new double[R];
    double* const mu_prime = new double[R];
    double logZ_prime = 0;

    /* if adaptive, load the per-example rate; otherwise, use the
       global rate */
    double lrate = (_adapt > 0 ? _lrates[xid] : _lrate);


    /* theoretically, it is always possible to find a learning rate
       that leads to improvement.  however, due to numerical error
       this can become impossible as the learning rate required for
       improvement may be small enough to cause underflow.  therefore,
       we must implement a hard limit on the number of cuts per
       example.  on the first pass, the limit is only 2 as a special
       case. */
    int num_remaining_cuts = (t == 1 ? 2 : EGSTRA_LRATE_MAXCUTS);
    while(true) { /* the loop breaks elsewhere */
      /* every iteration of this loop processes one example, whether a
	 cut is necessary or not */
      ++_num_processed;

      ////////////////////////////////////////
      // [1] - Calculate Theta Prime
      const double scale = lrate*_invNC;
      for(int r = 0; r < R; ++r) {
	theta_prime[r] =
	  Obj::update_dual_variable(theta[r], S[r], loss[r], scale);
      }

      ////////////////////////////////////////
      // [2] - Compute marginals prime with the thetas prime
      logZ_prime = _M->marginals(x, theta_prime, mu_prime);

      ////////////////////////////////////////
      // [3] - Compute the loss/entropy and the update to the primals
      /* create a sparse vector to hold the updates */
      sparsevec update;
      /* now, compute per-example dual and primal adjustment */
      double exdual_prime = Obj::exdual_init(logZ_prime);
      for(int r = 0; r < R; ++r) {
	exdual_prime +=
	  Obj::exdual_contrib(mu_prime[r], theta_prime[r], loss[r]);
	/* accumulate updates to primals */
	const double adjust = _invNC*(mu[r] - mu_prime[r]);
	parameters::sparse_add(update, F + r, adjust);
      }
      /* compute the change in the squared-L2 using the formula
	 (w + u)^2 = w^2 + 2wu + u^2
	 w^2 - (w + u)^2 = -2wu - u^2 */
      const double u_dot_w = _W.dot(update);
      const double u_sq = parameters::sparse_sqL2(update);
      const double sqL2_diff = -u_sq - 2*u_dot_w;
      /* the dual (minimization) improvement */
      const double dual_improvement =
	(exdual_cur - exdual_prime)/_N + 0.5*_C*sqL2_diff;

#ifdef REPORT_EACH_PROCESSED_EXAMPLE
      cerr << myname << " : t " << t << " s " << xid << " : "
	   << " len = " << len
	   << " nfeats = " << update.size()
	   << " lrate = " << lrate
	   << " e0 = " << exdual_cur << " e' = " << exdual_prime
	   << " u^2 = " << u_sq << " ndiff = " << sqL2_diff
	   << " impr = " << dual_improvement << flush;
#endif /* REPORT_EACH_PROCESSED_EXAMPLE */

      /* lack of improvement; cut the learning rate by half */
      if (dual_improvement < 0) {
	if(num_remaining_cuts > 0) {
#ifdef REPORT_EACH_PROCESSED_EXAMPLE
	  cerr << " [cut]" << endl;
#endif /* REPORT_EACH_PROCESSED_EXAMPLE */

	  --num_remaining_cuts; lrate *= 0.5;
	} else {                   /* exceeded hard limit on cuts */
	  /* terminate *without* updating the dual variables (no point
	     in setting the duals to demonstrably worse values). */
#ifdef REPORT_EACH_PROCESSED_EXAMPLE
	  cerr << " fail!" << endl;
#endif /* REPORT_EACH_PROCESSED_EXAMPLE */

	  /* discard the dual-prime variables */
	  delete [] theta_prime;
	  delete [] mu_prime;
	  /* don't save the old variables again */
	  _duals->discard(theta, mu);
	  /* do *not* store the learning rate.  see note [2] below */

	  break;
	}
      } else {     /* improvement; accept new duals/primals */
#ifdef REPORT_EACH_PROCESSED_EXAMPLE
	cerr << endl;
#endif /* REPORT_EACH_PROCESSED_EXAMPLE */

	/* commit the update to the primals */
	_W.add(update);

	if(_adapt > 0) {
	  /* store the learning rate that worked, with a small boost */
	  _lrates[xid] = lrate*_adapt;
	}
	/* update per-example dual */
	_exduals[xid] = exdual_prime;
	/* delete the current duals vectors */
	delete [] theta;
	delete [] mu;
	/* save the new duals for the round */
	_duals->set(xid, theta_prime, mu_prime, logZ_prime);

	break;
      }
    }

    /* clean up temporary data */
    free(loss);
    free(S);
  }


  template<class Obj>
  double deg_trainer<Obj>::estimate_lrate(const dataset& data) {
    //    srandom(time(NULL)); /* try to be random */
    srandom(0); /* repeatable behavior */
    assert(_N > 0);

    /* fix the behavior to NON-adaptive, and restore the proper
       behavior after the learning rate has been estimated */
    const double saved_adapt = _adapt;
    _adapt = -1;

    /* fixed fraction of the training data */
    const int sample_size = (int)(EGSTRA_LRATE_SAMPLERATIO*_N);
    cerr << myname << " : estimating learning rate from sample of "
	 << sample_size << " examples" << endl;
    assert(sample_size > 0);

    /* get sample_size unique random selections, using the prefix of a
       random permutation */
    dataset sample(data);
    for(int i = 0; i < sample_size; ++i) {
      optutils::randselect(EGSTRA_RAND_PERMUTATION, i, sample);
    }
    sample.resize(sample_size);

    /* we begin from an initial learning rate (provided by the
       objective function) and attempt to find a learning rate that
       results in a small number of cuts on this sample. */
    double lrate = 0.5*_N*_C * Obj::lrate_search_init();
    const int threshold_cuts =
      (int)(((double)sample_size)*EGSTRA_LRATE_CUTRATIO);
    while(1) { /* break within loop */
      const int cuts = count_cuts(sample, lrate);
      if(cuts > threshold_cuts) { lrate *= 0.5; }
      else                      { break; }
    }

    /* restore globals */
    _num_processed = 0;
    _adapt = saved_adapt;

    cerr << myname << " : estimated learning rate is "
	 << scientific << setprecision(16) << lrate << endl;
    return lrate;
  }


  /* perform *two* passes over the given sentences, using the given
     learning rate, and count the number of cuts on the second pass.
     the first or second iteration may be stopped early if the number
     of cuts becomes greater than nsents. */
  template<class Obj>
  int deg_trainer<Obj>::count_cuts(const dataset& sample,
				   const double lrate) {
    _lrate = lrate; /* NB : setting global */
    cerr << myname << " : trying lrate = " << _lrate << " : (init) " << flush;

    /* reinitialize the primal/dual parameters */
    _W.zero();
    assert(_duals == NULL);
    _duals = new egduals(_N, false);

    const int Nsample = sample.size();
    int num_cuts = 0;
    /* initialization pass */
    { /* enclose this in a block for the timer */
      timer_start();
      _num_processed = 0; /* reset the counter */
      for(int i = 0; i < Nsample; ++i) {
	const example* const ex = sample[i];
	const pattern* const x = ex->x();
	const struct fvec* const F = _fg->phi(x);
	initialize_example(ex, F);
	process_example(ex, F, 1);
	_fg->discard(F, x);
	const int num_examples = i + 1;
	num_cuts = _num_processed - num_examples;
	if(num_cuts >= Nsample) { break; } /* early stopping */
	ticker(num_examples, Nsample);
      }
      timer_stop();
      cerr << " [" << timestr(timediff) << "]" << flush;
      cerr << " " << num_cuts << flush;
      if(num_cuts >= Nsample) {
	cerr << endl;
	/* NB: unset the dual parameters */
	delete _duals;
	_duals = NULL;
	return num_cuts;
      }
    }

    /* actual cut-counting pass */
    cerr << "  (real) " << flush;
    _num_processed = 0; /* reset again */
    timer_start();
    for(int i = 0; i < Nsample; ++i) {
      const example* const ex = sample[i];
      const pattern* const x = ex->x();
      const struct fvec* const F = _fg->phi(x);
      process_example(ex, F, 2);
      _fg->discard(F, x);
      const int num_examples = i + 1;
      num_cuts = _num_processed - num_examples;
      if(num_cuts >= Nsample) { break; } /* early stopping */
      ticker(num_examples, Nsample);
    }
    _num_processed = 0; /* reset again */
    timer_stop();
    cerr << " [" << timestr(timediff) << "]" << flush;
    cerr << " " << num_cuts << " cuts" << endl;

    /* NB: unset the dual parameters */
    delete _duals;
    _duals = NULL;

    return num_cuts;
  }


  template <class Obj>
  void deg_trainer<Obj>::reconstruct_primals(const dataset& data) {
    cerr << myname << " : reconstructing primals : " << flush;
    timer_start();
    assert((int)data.size() == _N);
    const double old_norm = _W.sqL2(true);
    _W.zero();
    for(int i = 0; i < _N; ++i) {
      const example* const ex = data[i];
      const pattern* const x = ex->x();
      const int xid = x->id();
      const int R = x->nparts();

      /* create features for the current sentence */
      const struct fvec* const F = _fg->phi(x);

      /* retrieve the duals */
      const double* theta = NULL;
      const double* mu = NULL;
      double logZ = 0;
      _duals->get(xid, theta, mu, logZ);

      /* create vector of correct-part indicators */
      char* const correct = optutils::get_correct(ex);

      /* add contributions to primal:
	 W(A) = 1/C * \sum a_iy psi_iy */
      for(int r = 0; r < R; ++r) {
	_W.add(F + r, _invNC*(correct[r] - mu[r]));
      }

      _fg->discard(F, x);
      _duals->discard(theta, mu);
      free(correct);

      ticker(i + 1, _N);
    }
    ticker(_N, _N);
    timer_stop();
    cerr << " done [" << timestr(timediff) << "]" << endl;
    const double new_norm = _W.sqL2(true);
    const double diff = old_norm - new_norm;
    const double sum =  old_norm + new_norm;
    cerr << myname << scientific << setprecision(5)
	 << " : reconstructed norm diff = " << diff
	 << " (abs) " << 2.0*diff/sum << " (rel)" << endl;
  }
}

/* clean up these macros */
#undef myname
//#undef REPORT_EACH_PROCESSED_EXAMPLE

/* footnotes */

/* [1] i tried implementing a true pre-initialization, as per the
       formal definition of the algorithm, but this resulted in quite
       poor performance.  the rate of convergence was still fast, but
       the initial dual value (i.e., after the first pass) was several
       orders of magnitude larger than for on-the-fly initialization.
       the entropies are decoupled, so there wasn't a problem there;
       it was the norm of _W that was incredibly inflated.  when we
       initialize on-the-fly, _W starts out at zero.  when we visit an
       example for the first time, we add to _W an initial
       contribution that is derived from the marginals of the intial
       thetas, and then we immediately perform an update, which alters
       the contribution of this example to _W to more sensible values.
       in other words, _W always contains contributions that have been
       subject to at least one EG update.  when we pre-initialize,
       then _W gets filled with contributions from the initial duals,
       which may be quite senseless.  thus, the dual objective is
       greatly inflated and the EG algorithm has to spend the first
       few iterations bringing the inflated dual down to the level of
       the on-the-fly dual. */

/* [2] storing or not storing the learning rate is debatable.  i feel
       that improvement failure occurs because the duals are at a
       local optimum (when the other duals are held fixed).  thanks to
       the sampling and adaptivity, it's likely that the intial
       learning rate is not far off from a good value.  however, since
       the other duals are held fixed as we optimize each example,
       it's possible (and apparently somewhat frequent) for the duals
       on the current example to be locally optimal while being
       globally non-optimal.  however, as the other dual variables
       change, it may become possible to improve these duals in the
       future.  therefore, it is wiser to keep the learning rate at
       its initial level (so that meaningful improvement may be
       attained on the next time this example is visited) rather than
       store a learning rate that is reduced by many orders of
       magnitude */
